# Imports.

import numpy as np
from colormath.color_objects import sRGBColor, LabColor
from colormath.color_conversions import convert_color
import math

from .. import Global_Utilities as gu

# Function definitions.

def extract_colour_features( output_directory, file_data, data ):

    # L*.

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( data[0][i] )

    features = np.array( feature_1 )[:, np.newaxis]

    feature_names = ["Colour_L*"]

    # a*.

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( data[1][i] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Colour_a*" )

    # b*.

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( data[2][i] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Colour_b*" )

    # Delta E, difference from virgin.

    feature_1 = []

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data(file_data)

    virgin_mask = gu.produce_mask( sample_array, [16, 17, 19, 24] )

    virgin_L = data[0][virgin_mask]
    virgin_a = data[1][virgin_mask]
    virgin_b = data[1][virgin_mask]

    virgin_mean = [sum( data[0][virgin_mask] ) / len( data[0][virgin_mask] ), sum( data[1][virgin_mask] ) / len( data[1][virgin_mask] ), sum( data[2][virgin_mask] ) / len( data[2][virgin_mask] )]

    for i in range( len( file_data ) ):

        delta = math.sqrt( (virgin_mean[0] - data[0][i]) ** 2 + (virgin_mean[1] - data[1][i]) ** 2 + (virgin_mean[2] - data[2][i]) ** 2 )

        feature_1.append( delta )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Colour_DeltaE" )

    # Delta L, difference from virgin (but only L*).

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( abs( virgin_mean[0] - data[0][i] ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Colour_DeltaL" )

    df = gu.array_with_column_titles_to_df( features, feature_names )

    df.to_csv( output_directory + "Colour/Features/Features.csv" )

def read_and_analyse_features( ip, file_data ):

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "Colour/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        # ip.feature_selection = [i for i in range( len( feature_names ) )]
        ip.feature_selection = [0, 1, 2] # Paper 1. And current test features.

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Colour/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Colour/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Colour/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Colour/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "Colour/Features/Distance_Matrix.csv" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "Colour/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "Colour/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "Colour/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Colour/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "Colour/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Colour/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "Colour/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Colour/Features/", title = "Means_Dendrogram.pdf" )

def distance_to_virgin( file_data, data ):

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    array = np.hstack( (data[3][:, np.newaxis], data[4][:, np.newaxis], data[5][:, np.newaxis]) )

    gu.normalise_features( array )

    distance_matrix = gu.distance_matrix_from_features( array )

    distance = np.zeros( len( array ) )

    for i in range( len( array ) ):

        distance[i] = (distance_matrix[i][23] + distance_matrix[i][15] + distance_matrix[i][16] + distance_matrix[i][18]) / 4

    temp = distance.argsort()

    print( samples_present_array[temp] )

    print( distance[temp] )

    # [24 16 19 17 12  1 15 18  3  7  2  5  9  4 10 11  8  6 14 21 22 20 23 13]
    # [0.03974382 0.05419614 0.05511484 0.06760492 0.30163964 0.3595544
    #  0.45287626 0.54839772 0.57774539 0.58428028 0.58444907 0.63841646
    #  0.6426062  0.64468085 0.67112142 0.67640252 0.69095171 0.70618429
    #  0.72230624 0.79168456 0.82771439 0.84616013 0.88966182 1.89269532]

def convert_to_rgb( directory, output_directory, file_data, data ):

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    rgb_255 = []

    for i in range( len( samples_present ) ):

        L = data[3][i]
        a = data[4][i]
        b = data[5][i]

        lab = LabColor( *[L, a, b] )
        rgb = convert_color( lab, sRGBColor )
        rgb_255.append( [255 * color for color in rgb.get_value_tuple()] )

    rgb_255 = np.array( rgb_255 )

    rgb_255 = np.hstack( (np.array( [s for s in samples_present] )[:, np.newaxis], rgb_255) )

    np.savetxt( output_directory + "Colour/Sandbox/rgb_colours.csv", rgb_255, delimiter = "," )

def sandbox( directory, output_directory, file_data, data ):

    perform_distance_to_virgin = True

    perform_conversion_to_rgb = False

    if perform_distance_to_virgin:

        distance_to_virgin( file_data, data )

    if perform_conversion_to_rgb:

        convert_to_rgb( directory, output_directory, file_data, data )



# Imports

import pandas as pd
from rpy2.robjects import pandas2ri
from rpy2.robjects.conversion import localconverter
import rpy2.robjects as robjects

from . import Utilities as util

from .. import Global_Utilities as gu

# Function definitions.

def source_R_functions( directory ):

    r = robjects.r
    r['source']( directory + 'rHDPE_R_Functions.R' )

def authorise_googledrive( directory ):

    robjects.globalenv['authorise_googledrive']( directory )

def read_csv_from_gd_via_R( directory, filename ):

    read_csv_from_gd = robjects.globalenv['read_csv_from_gd']

    rdf = read_csv_from_gd( directory, filename )

    with localconverter( robjects.default_converter + pandas2ri.converter ):

        pydf = robjects.conversion.rpy2py( rdf )

    return pydf

def read_files_and_preprocess( ip, normalised = True, minimal_datasets = False, feature_1 = "", feature_2 = "" ):

    authorised = False

    dataset_names = ["FTIR", "DSC", "TGA", "Rheology", "TT", "Colour", "SHM", "TLS", "ESCR", "FTIR2", "FTIR3", "TGA_SB"]
    dataset_directories = ["FTIR/Features/", "DSC/Features/", "TGA/Features/", "Rheology/Features/", "TT/Features/", "Colour/Features/", "SHM/Features/", "TLS/Features/", "ESCR/Features/", "FTIR/Integral_Analysis/", "FTIR/Component_Analysis/Features/", "TGA/Sandbox/"]

    normalised_ext = ""

    if normalised == False:

        normalised_ext = "_Unnormalised"

    dataset, std_dataset = [], []

    for i in range( len( dataset_names ) ):

        if not minimal_datasets:

            if (i + 1) not in ip.datasets_to_read:

                continue

        n = dataset_names[i]

        if not normalised and (n == "FTIR2" or n == "FTIR3" or n == "TGA_SB"):

            continue

        try:

            df = pd.read_csv( ip.output_directory + dataset_directories[i] + "Mean_Features" + normalised_ext + ".csv", index_col = 0 )
            df_std = pd.read_csv( ip.output_directory + dataset_directories[i] + "Std_of_Features" + normalised_ext + ".csv", index_col = 0 )

        except FileNotFoundError:

            try:

                df = pd.read_csv( "Temp/" + dataset_directories[i] + "Mean_Features" + normalised_ext + ".csv", index_col = 0 )
                df_std = pd.read_csv( "Temp/" + dataset_directories[i] + "Std_of_Features" + normalised_ext + ".csv", index_col = 0 )

            except FileNotFoundError:

                if not authorised:

                    source_R_functions( ip.directory )

                    authorise_googledrive( ip.directory )

                    authorised = True

                print( "File not found on server, getting it from Google Drive." )

                df = read_csv_from_gd_via_R( dataset_directories[i], "~/Output/" + dataset_directories[i] + "Mean_Features" + normalised_ext + ".csv" )
                df_std = read_csv_from_gd_via_R( dataset_directories[i], "~/Output/" + dataset_directories[i] + "Std_of_Features" + normalised_ext + ".csv" )

        if minimal_datasets:

            if feature_1 not in df.columns.tolist():

                if feature_2 not in df.columns.tolist():

                    continue

        dataset.append( df )
        std_dataset.append( df_std )

    sample_mask = ip.sample_mask.copy()

    for i in range( len( dataset ) ):

        samples_present = dataset[i].iloc[:, 0].tolist()
        sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    features, feature_names = util.produce_full_dataset_of_features( dataset, sample_mask )

    rank_features = util.rank_features( features )

    features_df = gu.array_with_column_titles_and_label_titles_to_df( features, feature_names, sample_mask )

    rank_features_df = gu.array_with_column_titles_and_label_titles_to_df( rank_features, feature_names, sample_mask )

    std_of_features, _ = util.produce_full_dataset_of_features( std_dataset, sample_mask )

    std_of_features_df = gu.array_with_column_titles_and_label_titles_to_df( std_of_features, feature_names, sample_mask )

    #===============

    # Extracting the whole dataset as a .csv.

    # resin_data = gu.get_list_of_resins_data( ip.directory )
    #
    # sample_mask_2 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]
    #
    # features_2, feature_names_2, sample_mask_2 = util.compile_full_dataset_of_features( dataset, sample_mask_2 )
    # std_of_features_2, std_of_feature_names_2, sample_mask_2 = util.compile_full_dataset_of_features( std_dataset, sample_mask_2 )
    #
    # features_2_df = gu.array_with_column_titles_and_label_titles_to_df( features_2, feature_names_2, sample_mask_2 )
    # std_of_features_2_df = gu.array_with_column_titles_and_label_titles_to_df( std_of_features_2, std_of_feature_names_2, sample_mask_2 )
    #
    # features_2_df = features_2_df.set_axis( [resin_data.loc[i]["Label"] for i in features_2_df.index], axis = "index" )
    # std_of_features_2_df = std_of_features_2_df.set_axis( [resin_data.loc[i]["Label"] for i in std_of_features_2_df.index], axis = "index" )
    #
    # features_2_df.astype( float ).to_csv( ip.output_directory + "Global/Dataset/Full_Dataset.csv", float_format = "%.5f" )
    # std_of_features_2_df.astype( float ).to_csv( ip.output_directory + "Global/Dataset/Full_Dataset_Std.csv", float_format = "%.5f" )

    #===============

    return features_df, std_of_features_df, rank_features_df



# Imports.

import numpy as np
import matplotlib.pyplot as plt
from cycler import cycler

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, savefig = True ):

    resin_data = gu.get_list_of_resins_data( ip.directory )

    # For overall pipeline figure.

    # mpl.rcParams['lines.linewidth'] = 4

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    # samples_to_plot = samples_present[:-7]
    samples_to_plot = [23]

    specimens = False
    all_specimens = False
    specimen_mask_by_index = [1, 3]
    specimen_mask = []

    mean = True

    deriv0 = True
    deriv1 = False
    deriv2 = False

    cryst = True
    melt = True

    split = False
    num_splits = 15
    split_length = 10
    splits = [split_length * (i + 2) for i in range( num_splits )]
    splits = [130, 145]

    if not split:

        splits = [int( data[2][0] ), int( data[2][len( data[2] ) - 1] )]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        splits = ip.shiny_split
        melt = ip.shiny_melt
        cryst = ip.shiny_cryst
        mean = ip.shiny_mean
        specimens = ip.shiny_specimen
        all_specimens = False

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = True

    for s in range( len( splits ) - 1 ):

        data_extraction = []

        lower_bound, upper_bound = splits[s], splits[s + 1]

        for i in samples_to_plot:

            if specimens:

                mask = np.where( sample_array == i )[0]

                if not ip.shiny:

                    if all_specimens:

                        specimen_mask = [file_data[mask[j]][2] for j in range( len( mask ) )]

                    else:

                        specimen_mask = [file_data[mask[j]][2] for j in specimen_mask_by_index]

                for j in mask:

                    if file_data[j][2] in specimen_mask:

                        if deriv0:

                            if cryst:

                                temp_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                                plt.plot( data[0][temp_mask], data[1][j][temp_mask], label = file_data[j][2] )

                                shiny_de.append( data[0][temp_mask].tolist() )
                                shiny_de.append( data[1][j][temp_mask].tolist() )

                            if melt:

                                temp_mask = np.where( (data[2] <= upper_bound) & (data[2] >= lower_bound) )[0]

                                plt.plot( data[2][temp_mask], data[3][j][temp_mask], label = file_data[j][2] )

                                shiny_de.append( data[2][temp_mask].tolist() )
                                shiny_de.append( data[3][j][temp_mask].tolist() )

                        if deriv1:

                            if cryst:

                                temp_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                                plt.plot( first_derivative_data[0][temp_mask], first_derivative_data[1][j][temp_mask], label = file_data[j][2] )

                            if melt:

                                temp_mask = np.where( (first_derivative_data[2] <= upper_bound) & (first_derivative_data[2] >= lower_bound) )[0]

                                plt.plot( first_derivative_data[2][temp_mask], first_derivative_data[3][j][temp_mask], label = file_data[j][2] )

                        if deriv2:

                            if cryst:

                                temp_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                                plt.plot( second_derivative_data[0][temp_mask], second_derivative_data[1][j][temp_mask], label = file_data[j][2] )

                            if melt:

                                temp_mask = np.where( (second_derivative_data[2] <= upper_bound) & (second_derivative_data[2] >= lower_bound) )[0]

                                plt.plot( second_derivative_data[2][temp_mask], second_derivative_data[3][j][temp_mask], label = file_data[j][2] )

            if mean:

                index = np.where( samples_present_array == i )[0][0]

                if deriv0:

                    if cryst:

                        temp_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                        plt.plot( data[0][temp_mask], data[4][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                        shiny_de.append( data[0][temp_mask].tolist() )
                        shiny_de.append( data[4][index][temp_mask].tolist() )

                        data_extraction.append( data[0][temp_mask] )
                        data_extraction.append( data[4][index][temp_mask] )

                    if melt:

                        temp_mask = np.where( (data[2] <= upper_bound) & (data[2] >= lower_bound) )[0]

                        plt.plot( data[2][temp_mask], data[5][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                        shiny_de.append( data[2][temp_mask].tolist() )
                        shiny_de.append( data[5][index][temp_mask].tolist() )

                        data_extraction.append( data[2][temp_mask] )
                        data_extraction.append( data[5][index][temp_mask] )

                if deriv1:

                    if cryst:

                        temp_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                        plt.plot( first_derivative_data[0][temp_mask], first_derivative_data[4][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    if melt:

                        temp_mask = np.where( (first_derivative_data[2] <= upper_bound) & (first_derivative_data[2] >= lower_bound) )[0]

                        plt.plot( first_derivative_data[2][temp_mask], first_derivative_data[5][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                if deriv2:

                    if cryst:

                        temp_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                        plt.plot( second_derivative_data[0][temp_mask], second_derivative_data[4][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    if melt:

                        temp_mask = np.where( (second_derivative_data[2] <= upper_bound) & (second_derivative_data[2] >= lower_bound) )[0]

                        plt.plot( second_derivative_data[2][temp_mask], second_derivative_data[5][index][temp_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

        if ip.shiny:

            return shiny_de

        # plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )
        # plt.legend()

        plt.xlabel( "Temperature °C" )
        plt.ylabel( "Heat Flow" )

        plt.tight_layout()

        # For overall pipeline figure.

        # ax = plt.gca()
        # ax.get_legend().remove()
        # plt.xlabel( "" )
        # plt.ylabel( "" )
        # plt.tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        # plt.tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        if savefig:

            plt.savefig( ip.output_directory + "DSC/Plots/Plot.pdf" )

        else:

            plt.show()

        plt.close()

        if data_extraction_bool:

            data_extraction[0] = np.pad( data_extraction[0], (0, 500) )
            data_extraction[1] = np.pad( data_extraction[1], (0, 500) )

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

        return shiny_de

def plot_variance( directory, output_directory, data, mask, s, std_c, std_m ):

    resin_data = gu.get_list_of_resins_data( directory )

    plt.gca().set_prop_cycle( cycler( color = ['c', 'c', 'm', 'm', 'y', 'y', 'k', 'k', 'g', 'g'] ) )

    for i in mask:

        plt.plot( data[0], data[1][i], linewidth = 1 )
        plt.plot( data[2], data[3][i], linewidth = 1, label = i )

    x_axis_1 = np.linspace( 0, 0, len( data[0] ) )
    x_axis_2 = np.linspace( 0, 0, len( data[2] ) )

    plt.gca().set_prop_cycle( cycler( color = ['c', 'm'] ) )

    plt.fill_between( data[0], std_c, x_axis_1, linewidth = 1.5 )
    plt.fill_between( data[2], -np.array( std_m ), x_axis_2, linewidth = 1.5 )

    plt.title( resin_data.loc[s]["Label"] )
    plt.xlabel( "Temperature (°C)" )
    plt.ylabel( "Normalised Heat Flow" )
    plt.legend()
    plt.savefig( output_directory + resin_data.loc[s]["Label"] + ".pdf" )
    plt.close()



# Imports.

import numpy as np
import pandas as pd
from scipy.stats import pearsonr

from .. import Global_Utilities as gu

# Function definitions.

def compute_derivatives( data, width = 1 ):
    '''Compute the derivatives.'''

    derivative_data = [data[0][width: -width], [], [], []]

    for i in range( len( data[1] ) ):

        derivative_data[1].append( np.array( gu.derivative( data[0], data[1][i], width ) ) )

    for i in range( len( data[3] ) ):

        derivative_data[3].append( np.array( gu.derivative( data[0], data[3][i], width ) ) )

    return derivative_data

def extract_GCMS_features( output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data ):

    pass

def read_and_analyse_FTIR_features( directory, ip, file_data ):

    pass

def sandbox( directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data ):

    pass



# Imports.

import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def compute_derivatives( data, width = 1 ):
    '''Compute the derivatives.'''

    derivative_data = [data[0][width:-width], [], [], [], [], [], [], [], []]

    for i in range( len( data[3] ) ):

        derivative_data[3].append( np.array( gu.derivative( data[0], data[3][i], width ) ) )

    for i in range( len( data[7] ) ):

        derivative_data[7].append( np.array( gu.derivative( data[0], data[7][i], width ) ) )

    return derivative_data

def extract_rheology_features( output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    # Viscosities at all angular frequencies.

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( np.log10( data[3][i][0] ) )

    features = np.array( feature_1 )[:, np.newaxis]

    feature_names = ["Rhe_{:.2f}".format( data[0][0] )]

    for j in range( 1, len( data[0] ) ):

        feature_1 = []

        for i in range( len( file_data ) ):

            feature_1.append( np.log10( data[3][i][j] ) )

        features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

        feature_names.append( "Rhe_{:.2f}".format( data[0][j] ) )

    # SM at all angular frequencies

    for j in range( len( data[1][1] ) ):

        feature_1 = []

        for i in range( len( file_data ) ):

            feature_1.append( np.log10( data[1][i][j] ) )

        features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

        feature_names.append( "Rhe_SM_{:.2f}".format( data[0][j] ) )

    # LM at all angular frequencies

    for j in range( len( data[2][1] ) ):

        feature_1 = []

        for i in range( len( file_data ) ):

            feature_1.append( np.log10( data[2][i][j] ) )

        features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

        feature_names.append( "Rhe_LM_{:.2f}".format( data[0][j] ) )

    # Loss factor at all angular frequencies

    for j in range( len( data[4][1] ) ):

        feature_1 = []

        for i in range( len( file_data ) ):

            feature_1.append( data[4][i][j] )

        features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

        feature_names.append( "Rhe_Loss_{:.2f}".format( data[0][j] ) )

    # Crossover point.

    feature_1, feature_2 = [], []

    for i in range( len( file_data ) ):

        index_1, index_2 = -2, -1

        for j in range( len( data[4][i] ) ):

            if data[4][i][j] > 1:

                index_1 = j - 1
                index_2 = j
                break

        log_change = (np.log10( 1 ) - np.log10( data[4][i][index_1] )) / (np.log10( data[4][i][index_2] ) - np.log10( data[4][i][index_1] ))

        crossover_point = 10 ** ((np.log10( data[0][index_2] ) - np.log10( data[0][index_1] )) * log_change + np.log10( data[0][index_1] ))

        storage_modulus_crossover_point = 10 ** ((np.log10( data[1][i][index_2] ) - np.log10( data[1][i][index_1] )) * log_change + np.log10( data[1][i][index_1] ))

        feature_1.append( np.log10( crossover_point ) )
        feature_2.append( np.log10( storage_modulus_crossover_point ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )
    features = np.hstack( (features, np.array( feature_2 )[:, np.newaxis]) )

    feature_names.append( "Rhe_Crossover" )
    feature_names.append( "Rhe_SMCrossover" )

    # Log difference between extrapolated line of zero shear and reality.

    feature_1 = []

    for i in range( len( file_data ) ):

        m = (np.log( data[3][i][3] ) - np.log( data[3][i][1] )) / (np.log( data[0][3] ) - np.log( data[0][1] ))
        b = np.log( data[3][i][2] / data[0][2] ** m )

        extrapolated_value = data[0][len( data[0] ) - 1] ** m * np.exp( b )
        real_value = data[3][i][len( data[3][i] ) - 1]

        feature_1.append( np.log( extrapolated_value ) - np.log( real_value ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Rhe_Log" )

    # V_deg value.

    feature_1 = []

    for i in range( len( file_data ) ):

        deriv = gu.derivative( np.log( data[3][i] * data[0] ), np.arctan( data[4][i] ) )

        feature_1.append( (deriv[14] - deriv[len( deriv ) - 5]) / (data[3][i][len( data[3][i] ) - 1] * data[0][len( data[0] ) - 1]) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Rhe_V_deg" )

    # Custom V_deg value.

    feature_1 = []

    for i in range( len( file_data ) ):

        deriv = gu.derivative( np.log( data[3][i] ), data[4][i] )

        feature_1.append( deriv[14] - deriv[len( deriv ) - 5] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "Rhe_V_deg_custom" )

    df = gu.array_with_column_titles_to_df( features, feature_names )

    df.to_csv( output_directory + "Rheology/Features/Features.csv" )

def read_and_analyse_features( ip, file_data ):

    add_Carreau_Yasuda = False

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "Rheology/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]
        # ip.sample_mask = [16, 17, 19, 24, 12, 15, 1, 2, 3, 5, 6, 7, 8, 9, 4, 10, 13, 11, 14, 18, 20, 21, 22, 23]

        if add_Carreau_Yasuda:

            ip.sample_mask.pop( 23 )
            ip.sample_mask.pop( 22 )
            ip.sample_mask.pop( 6 )

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        # ip.feature_selection = [i for i in range( len( feature_names ) )]
        ip.feature_selection = [0, 15, 30, 31, 46, 61, 62, 77, 92, 93, 108, 123, 124, 126]
        # ip.feature_selection = [0, 15, 30, 124, 126]
        # ip.feature_selection = [127, 128]
        # ip.feature_selection = [0, 30, 93, 123, 124, 125, 126]
        # ip.feature_selection = [0, 93, 30, 123, 126, 124, 125]
        # ip.feature_selection = [i for i in range( 62 )]

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Rheology/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Rheology/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features, feature_weights )

    # Combine similar features and add a weighting.

    feature_combinations = [[0, 3, 6, 9], [1, 4, 7, 10], [2, 5, 8, 11], [12], [13]]

    features, feature_names = gu.combine_features( features, feature_names, feature_combinations )

    feature_weights = [2/3, 2/3, 2/3, 1, 1]

    gu.normalise_features( features, feature_weights )

    #####

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    if add_Carreau_Yasuda:

        # Add Carreau-Yasuda Data (No V1 data).

        eta_zero = [1, 1.71, 1.26, 2.37, 0.127, 1.25, 1.15, 0.798, 15.4, 2.77, 0.753, 0.297, 9.36, 1.19, 0.289, 0.752, 0.133, 7.22, 1183000, 0.0885, 12.6, 5140, 21.3, 402, 1]
        eta_zero = [np.log10( i ) for i in eta_zero]

        mean_features, mean_feature_names = gu.add_mean_feature( mean_features, mean_feature_names, sample_mask, eta_zero, "Rhe_Eta_Zero", weight = 1 )

        eta_inf = [1, 87774, 98324, 115000, 106000, 105000, 76214, 98396, 80000, 54242, 94807, 147000, 97278, 111000, 166000, 116000, 239000, 82292, 37074, 276000, 68147, 12401, 32037, 40000, 80000] # PCR 8 and 23 artificial data.
        eta_inf = [np.log10( i ) for i in eta_inf]

        mean_features, mean_feature_names = gu.add_mean_feature( mean_features, mean_feature_names, sample_mask, eta_inf, "Rhe_Eta_Inf", weight = 1 )

        lambda_param = [0, 0.13594, 0.20048, 0.06788, 0.011394, 0.23778, 0.14816, 0.16169, 0.94035, 0.21901, 0.11131, 0.090291, 0.44758, 0.17766, 0.10097, 0.038893, 0.12605, 0.07489, 0.085517, 0.1136, 0.12143, 0.72155, 0.10464, 0.083242, 0.1]

        mean_features, mean_feature_names = gu.add_mean_feature( mean_features, mean_feature_names, sample_mask, lambda_param, "Rhe_Lambda", weight = 1 )

        a_param = [0, 0.12552, 0.13952, 0.10886, 0.15203, 0.13854, 0.14265, 0.1569, 0.12139, 0.12649, 0.14568, 0.1725, 0.10818, 0.14222, 0.17806, 0.12853, 0.26823, 0.09261, 0.039129, 0.29079, 0.090029, 0.055225, 0.084577, 0.06129, 0.1]

        mean_features, mean_feature_names = gu.add_mean_feature( mean_features, mean_feature_names, sample_mask, a_param, "Rhe_A", weight = 1 )

        feature_combinations = [[0], [1], [2], [3], [4, 6, 8], [5], [7]]

        mean_features, mean_feature_names = gu.combine_features( mean_features, mean_feature_names, feature_combinations )

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    if add_Carreau_Yasuda:

        std_of_features = np.hstack( (std_of_features, np.zeros( len( sample_mask ) )[:, np.newaxis]) )
        std_of_features = np.hstack( (std_of_features, np.zeros( len( sample_mask ) )[:, np.newaxis]) )
        std_of_features = np.hstack( (std_of_features, np.zeros( len( sample_mask ) )[:, np.newaxis]) )
        std_of_features = np.hstack( (std_of_features, np.zeros( len( sample_mask ) )[:, np.newaxis]) )

        std_of_features, _ = gu.combine_features( std_of_features, feature_combinations = feature_combinations )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Rheology/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "Rheology/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "Rheology/Features/Distance_Matrix.csv" )

    # Perform PCA for rheology features.

    # gu.perform_pca( ip.directory, gu.array_with_column_titles_to_df( mean_features, ["Rhe_Feature_" + str( i + 1 ) for i in range( len( mean_feature_names ) )] ), sample_mask, std_error = True, std_of_features_df = pd.DataFrame( std_of_features ), num_components = 2, filename = ip.output_directory + "Rheology/Features/PCA.pdf" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "Rheology/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "Rheology/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "Rheology/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Rheology/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "Rheology/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Rheology/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "Rheology/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "Rheology/Features/", title = "Means_Dendrogram.pdf" )

def sandbox( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    pass



# Imports

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    with open( filename, 'r' ) as file:

        lines = file.readlines()

        a_list = lines[0].split( "," )

        data[0].append( float( a_list[0] ) )
        data[1].append( float( a_list[1] ) )
        data[2].append( float( a_list[2] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], [], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_2 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    data[0] = np.array( data[0] )
    data[1] = np.array( data[1] )
    data[2] = np.array( data[2] )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    pass

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    pass

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    pass

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv files.'''

    return [], []

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    # files_to_remove = []
    #
    # for i in range( len( file_data ) ):
    #
    #     s = file_data[i][3]
    #
    #     for l in descriptors_to_remove:
    #
    #         if s.find( l ) > -0.5:
    #
    #             files_to_remove.append( i )
    #
    #             break
    #
    # files_to_remove.reverse()
    #
    # for r in files_to_remove:
    #
    #     file_data.pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    mL = [i[0] for i in gu.sample_mean( file_data, [[data[0][j]] for j in range( len( file_data ) )] )]
    ma = [i[0] for i in gu.sample_mean( file_data, [[data[1][j]] for j in range( len( file_data ) )] )]
    mb = [i[0] for i in gu.sample_mean( file_data, [[data[2][j]] for j in range( len( file_data ) )] )]

    m = [np.array( mL ), np.array( ma ), np.array( mb )]

    labels = ["L", "a", "b"]

    for i in range( len( m ) ):

        np.savetxt( output_directory + "Colour/Mean_Data/" + labels[i] + "_Means" + name_appendage + ".csv", m[i][np.newaxis, :], delimiter = ",", fmt = "%.3f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    labels = ["L", "a", "b"]

    for i in range( len( labels ) ):

        m = []

        df = pd.read_csv( output_directory + "Colour/Mean_Data/" + labels[i] + "_Means" + name_appendage + ".csv", sep = ",", header = None )

        for j in range( len( df.columns ) ):

            m.append( df.iloc[0, j] )

        if len( data ) > 3 + i: # A previous list of means is already appended.

            data[3 + i] = np.array( m )

        else:

            data.append( np.array( m ) )

    return data



# Imports.

import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def extract_ESCR_features( output_directory, file_data, data ):

    # ESCR_90%Failure.

    feature_1 = []

    for i in range( len( file_data ) ):

        mask = np.where( data[1][i] <= 10 )[0]

        if len( mask ) == 0:

            feature_1.append( 96 )
            continue

        if file_data[i][0] == 3 or file_data[i][0] == 9:

            if data[0][mask[0]] == 48:

                hours_1, hours_2 = 24, 48

            else:

                hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        else:

            hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        percentage_1, percentage_2 = data[1][i][mask[0] - 1], data[1][i][mask[0]]

        below_10 = 10 - percentage_2
        percentage_change = percentage_1 - percentage_2
        hour_change = hours_2 - hours_1

        feature_1.append( hours_1 + hour_change * (1 - below_10 / percentage_change) )

    features = np.array( feature_1 )[:, np.newaxis]

    feature_names = ["ESCR_90%Failure"]

    # ESCR_75%Failure.

    feature_1 = []

    for i in range( len( file_data ) ):

        mask = np.where( data[1][i] <= 25 )[0]

        if len( mask ) == 0:

            feature_1.append( 96 )
            continue

        if file_data[i][0] == 3 or file_data[i][0] == 9:

            if data[0][mask[0]] == 48:

                hours_1, hours_2 = 24, 48

            else:

                hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        else:

            hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        percentage_1, percentage_2 = data[1][i][mask[0] - 1], data[1][i][mask[0]]

        below_25 = 25 - percentage_2
        percentage_change = percentage_1 - percentage_2
        hour_change = hours_2 - hours_1

        feature_1.append( hours_1 + hour_change * (1 - below_25 / percentage_change) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "ESCR_75%Failure" )

    # ESCR_50%Failure.

    feature_1 = []

    for i in range( len( file_data ) ):

        mask = np.where( data[1][i] <= 50 )[0]

        if len( mask ) == 0:

            feature_1.append( 96 )
            continue

        if file_data[i][0] == 3 or file_data[i][0] == 9:

            if data[0][mask[0]] == 48:

                hours_1, hours_2 = 24, 48

            else:

                hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        else:

            hours_1, hours_2 = data[0][mask[0] - 1], data[0][mask[0]]

        percentage_1, percentage_2 = data[1][i][mask[0] - 1], data[1][i][mask[0]]

        below_50 = 50 - percentage_2
        percentage_change = percentage_1 - percentage_2
        hour_change = hours_2 - hours_1

        feature_1.append( hours_1 + hour_change * (1 - below_50 / percentage_change) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "ESCR_50%Failure" )

    df = gu.array_with_column_titles_to_df( features, feature_names )

    df.to_csv( output_directory + "ESCR/Features/Features.csv" )

def read_and_analyse_ESCR_features( ip, file_data ):

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "ESCR/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        ip.feature_selection = [i for i in range( len( feature_names ) )]

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "ESCR/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "ESCR/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "ESCR/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "ESCR/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "ESCR/Features/Distance_Matrix.csv" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "ESCR/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "ESCR/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "ESCR/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "ESCR/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "ESCR/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "ESCR/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "ESCR/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "ESCR/Features/", title = "Means_Dendrogram.pdf" )

def sandbox( directory, output_directory, file_data, data ):

    pass



# Imports.

from numbers_parser import Document

# Class definitions.

class Input_Parameters():

    def __init__( self ):

        self.shiny = False

        self.directory = ""

        self.data_directory = ""

        self.output_directory = ""

        self.read_files = False

        self.merge_groups = False

        self.write_csv = False

        self.read_csv = False

        self.remove_files = False

        self.compute_mean = False

        self.read_mean = False

        self.derivative = False

        self.extract_features = False

        self.read_and_analyse_features = False

        self.plot_data = False

        self.sandbox = False

        self.sample_mask = []

        self.feature_selection = []

        self.shiny_samples_to_plot = []

        self.shiny_specimens_to_plot = []

        self.shiny_split = []

        self.shiny_melt = False

        self.shiny_cryst = False

        self.shiny_mean = False

        self.shiny_specimen = False

        self.name_appendage = ""

# Function definitions.

def read_parameters_from_numbers_file( filename, ip ):

    doc = Document( filename )

    sheets = doc.sheets
    tables = sheets[0].tables
    rows = tables[0].rows()

    ip.shiny = bool( rows[0][1].value )

    ip.directory = str( rows[1][1].value or "" )

    ip.data_directory = str( rows[2][1].value or "" )

    ip.output_directory = str( rows[3][1].value or "" )

    ip.read_files = bool( rows[4][1].value )

    ip.merge_groups = bool( rows[5][1].value )

    ip.write_csv = bool( rows[6][1].value )

    ip.read_csv = bool( rows[7][1].value )

    ip.remove_files = bool( rows[8][1].value )

    ip.compute_mean = bool( rows[9][1].value )

    ip.read_mean = bool( rows[10][1].value )

    ip.derivative = bool( rows[11][1].value )

    ip.extract_features = bool( rows[12][1].value )

    ip.read_and_analyse_features = bool( rows[13][1].value )

    ip.plot_data = bool( rows[14][1].value )

    ip.sandbox = bool( rows[15][1].value )

    ip.sample_mask = [int( i.value ) for i in rows[16][1:] if i.value != None]

    ip.feature_selection = [int( i.value ) for i in rows[17][1:] if i.value != None]



# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_until_break( f, T, Temp, HF, gap ):
    '''Read the file until there is a break.'''

    linenumber = 0

    time = []
    temperature = []
    heat_flow = []

    for line in f:

        if linenumber >= gap:

            if line.rstrip():

                a_list = line.rstrip().split( ',' )

                if a_list[0] and a_list[1] and a_list[2]:

                    map_object = list( map( float, a_list ) )

                    time.append( map_object[0] )
                    temperature.append( map_object[1] )
                    heat_flow.append( map_object[2] )

            else:

                break

        linenumber += 1

    T.append( time )
    Temp.append( temperature )
    HF.append( heat_flow )

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    Time = [[] for i in range( 6 )]
    Temperature = [[] for i in range( 6 )]
    Heat_Flow = [[] for i in range( 6 )]

    with open( filename, 'r' ) as f:

        read_until_break( f, Time[0], Temperature[0], Heat_Flow[0], 10 )

        read_until_break( f, Time[1], Temperature[1], Heat_Flow[1], 4 )

        read_until_break( f, Time[2], Temperature[2], Heat_Flow[2], 4 )

        read_until_break( f, Time[3], Temperature[3], Heat_Flow[3], 4 )

        read_until_break( f, Time[4], Temperature[4], Heat_Flow[4], 4 )

        read_until_break( f, Time[5], Temperature[5], Heat_Flow[5], 4 )

    data[0].append( np.array( Temperature[2][0] )  )
    data[1].append( np.array( Heat_Flow[2][0] ) )
    data[2].append( np.array( Temperature[4][0] ) )
    data[3].append( np.array( Heat_Flow[4][0] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def read_raw_data_file_2( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_(\D)" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    phase = pattern.search( f ).groups()[2]

    Time = [[] for i in range( 6 )]
    Temperature = [[] for i in range( 6 )]
    Heat_Flow = [[] for i in range( 6 )]

    if phase == "C":

        with open( filename, 'r' ) as f:

            read_until_break( f, Time[2], Temperature[2], Heat_Flow[2], 3 )

        data[0].append( np.array( Temperature[2][0] ) )
        data[1].append( np.array( Heat_Flow[2][0] ) )

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    else:

        with open( filename, 'r' ) as f:

            read_until_break( f, Time[4], Temperature[4], Heat_Flow[4], 3 )

        data[2].append( np.array( Temperature[4][0] ) )
        data[3].append( np.array( Heat_Flow[4][0] ) )

def read_raw_data_file_3( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    Temperature = [[] for i in range( 6 )]
    Heat_Flow = [[] for i in range( 6 )]

    with open( filename, 'r', encoding = "windows-1252" ) as f:

        linenumber = 0

        temp = []
        heat_flow = []

        for line in f:

            if linenumber < 3:

                linenumber += 1
                continue

            line_data = line.split( "," )

            temp.append( float( line_data[0] ) )
            heat_flow.append( float( line_data[1] ) )

            if abs( temp[ len( temp ) - 1] - temp[ len( temp ) - 2] ) > 1:

                break

        temp = []
        heat_flow = []

        for line in f:

            line_data = line.split( "," )

            temp.append( float( line_data[0] ) )
            heat_flow.append( float( line_data[1] ) )

            if abs( temp[ len( temp ) - 1] - temp[ len( temp ) - 2] ) > 1:

                break

        Temperature[2].append( temp )
        Heat_Flow[2].append( heat_flow )

        temp = []
        heat_flow = []

        for line in f:

            line_data = line.split( "," )

            temp.append( float( line_data[0] ) )
            heat_flow.append( float( line_data[1] ) )

            if abs( temp[ len( temp ) - 1] - temp[ len( temp ) - 2] ) > 1:

                break

        Temperature[4].append( temp )
        Heat_Flow[4].append( heat_flow )

    data[0].append( np.array( Temperature[2][0] ) )
    data[1].append( np.array( Heat_Flow[2][0] ) )
    data[2].append( np.array( Temperature[4][0] ) )
    data[3].append( np.array( Heat_Flow[4][0] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], [], [], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_2 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            if resin == 28:

                read_raw_data_file_2( data_directory + r + "/" + f, f, resin_data, file_data, data )

            elif resin == 40 or resin == 41:

                read_raw_data_file_3( data_directory + r + "/" + f, f, resin_data, file_data, data )

            else:

                read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    standard_temp_list = np.linspace( 55, 165, 5500 ).tolist() # 195, 7000
    reverse_standard_temp_list = standard_temp_list[::-1]

    for i in range( len( data[0] ) ):

        mask = []

        for j in range( len( reverse_standard_temp_list ) ):

            interval_mask = np.where( data[0][i] <= reverse_standard_temp_list[j] )[0]

            if interval_mask.any():

                mask.append( interval_mask[0] )

            else:

                mask.append( len( data[0][i] ) - 1 )

        data[1][i] = data[1][i][mask] - data[1][i][mask].min()

        print( i )

    standard_temp_list = np.linspace( 55, 175, 6000 ).tolist()

    for i in range( len( data[2] ) ):

        mask = []

        for j in range( len( standard_temp_list ) ):

            interval_mask = np.where( data[2][i] >= standard_temp_list[j] )[0]

            if interval_mask.any():

                mask.append( interval_mask[0] )

            else:

                mask.append( len( data[2][i] ) - 1 )

        data[3][i] = data[3][i][mask] - data[3][i][mask].max()

        print( i )

    data[0] = np.array( reverse_standard_temp_list )
    data[2] = np.array( standard_temp_list )

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    # Add a for specimens that may be anomalous.
    # Add d for specimens that are duplicates.
    # Add u for unusual specimens.
    # Add z for PP specimens.
    # Add p for specimens for which the DSC was performed by Peiyao.

    specimens = {13:[0], 18:[0]} # Was 13:[0], 14:[2], 16:[0], 18:[0], 19:[1], 21:[2], 24:[2], changed on 26/6/24.

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "a"

    specimens = {23:[2, 4], 1:[2], 9:[2], 110:[2]} # 9:[2] moved from "a" to "d", 1:[2] and 110:[2] added, 26/6/24.

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "d"

    specimens = {24:[0]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "u"

    specimens = {201:[0, 1]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "z"

    for f in file_data:

        if f[0] >= 101 and f[0] <= 200:

            f[3] = f[3] + "p"

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    array = data[0][:, np.newaxis]

    for i in range( len( data[1] ) ):

        array = np.hstack( (array, data[1][i][:, np.newaxis]) )

    np.savetxt( output_directory + "DSC/Condensed_Data/Crystallisation" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.3f" )

    array = data[2][:, np.newaxis]

    for i in range( len( data[3] ) ):

        array = np.hstack( (array, data[3][i][:, np.newaxis]) )

    np.savetxt( output_directory + "DSC/Condensed_Data/Melt" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.3f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "DSC/Condensed_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    file_data = []

    df = pd.read_csv( output_directory + "DSC/Condensed_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = df.iloc[i, 0]
        specimen = df.iloc[i, 1]

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    data = []

    df = pd.read_csv( output_directory + "DSC/Condensed_Data/Crystallisation" + name_appendage + ".csv", sep = ",", header = None )

    data.append( df.iloc[:, 0].to_numpy( dtype = np.float64 ) )

    Heat_Flow = []

    for i in range( 1, len( df.columns ) ):

        Heat_Flow.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    data.append( Heat_Flow )

    df = pd.read_csv( output_directory + "DSC/Condensed_Data/Melt" + name_appendage + ".csv", sep = ",", header = None )

    data.append( df.iloc[:, 0].to_numpy( dtype = np.float64 ) )

    Heat_Flow = []

    for i in range( 1, len( df.columns ) ):

        Heat_Flow.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    data.append( Heat_Flow )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[1].pop( r )
        data[3].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    m = [gu.sample_mean( file_data, data[1] ), gu.sample_mean( file_data, data[3] )]

    labels = ["Crystallisation", "Melt"]

    for i in range( len( m ) ):

        array = m[i][0][:, np.newaxis]

        for j in range( 1, len( m[i] ) ):

            array = np.hstack( (array, m[i][j][:, np.newaxis]) )

        np.savetxt( output_directory + "DSC/Condensed_Data/" + labels[i] + "_Means" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.4f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    labels = ["Crystallisation", "Melt"]

    for i in range( len( labels ) ):

        m = []

        df = pd.read_csv( output_directory + "DSC/Condensed_Data/" + labels[i] + "_Means" + name_appendage + ".csv", sep = ",", header = None )

        for j in range( len( df.columns ) ):

            m.append( df.iloc[:, j].to_numpy( dtype = np.float64 ) )

        if len( data ) > 4 + i: # A previous list of means is already appended.

            data[4 + i] = m

        else:

            data.append( m )

    return data



if bound_temperature:

    Preprocessing.bound_temperature( data, m, lower_bound, upper_bound )

def bound_range( temperature, lower_bound, upper_bound ):
    '''Creates a mask with which to bound the data.'''

    mask = np.where( (temperature <= upper_bound) & (temperature >= lower_bound) )[0]

    return mask

def bound_temperature( data, m, lower_bound, upper_bound ):
    '''Bound the data so that it lies within a specified temperature range.'''

    range_mask = bound_range( np.array( data[1][0] ), lower_bound, upper_bound )

    for i in range( len( data[3] ) ):

        data[3][i] = list( np.array( data[3][i] )[range_mask] )

    data[1][0] = list( np.array( data[1][0] )[range_mask] )

    for i in range( len( m ) ):

        m[i] = list( np.array( m[i] )[range_mask] )

def deriv_simplification( y, width ):

    deriv_simp = []

    for i in range( width, len( y ) - width ):

        list = y[i - width:i + width]
        array = np.array( list )
        min = array.min()
        max = array.max()
        deriv_simp.append( 0.5 * (min + max) )

    return deriv_simp

# Mean derivative at a selection of 20 degree ranges using deriv_simplification.

for j in range( 28 ):

    feature_1 = []

    # if j not in [6, 9, 13, 15, 17, 19, 21]:
    #
    #     continue

    lower_bound = 20 * (j + 2)
    upper_bound = 20 * (j + 3)

    mask = np.where( (np.array( data[1][0] ) >= lower_bound) & (np.array( data[1][0] ) <= upper_bound) )

    for i in range( len( file_data ) ):

        x = list( np.array( data[1][i] )[mask] )
        y = list( np.array( data[3][i] )[mask] )

        width = 7
        simp = 12
        step = 1

        y = gu.derivative( x, y, width )
        y = deriv_simplification( y, simp )

        feature_1.append( np.mean( np.array( y ) ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "TGA_Temp_{}-{}".format( lower_bound, upper_bound ) )



# Imports.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, savefig = False ):

    resin_data = gu.get_list_of_resins_data( ip.directory )

    # For overall pipeline figure.

    # mpl.rcParams['lines.linewidth'] = 4

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    # samples_to_plot = samples_present
    # samples_to_plot = [1, 3, 6, 8, 9, 12, 15, 17, 24, 25]
    samples_to_plot = [17, 24, 25]

    split = False

    if not split:

        splits = [0, 96]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        splits = ip.shiny_split

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = False

    plt.figure( figsize = (10, 8) )

    for s in range( len( splits ) - 1 ):

        data_extraction = []

        lower_bound, upper_bound = splits[s], splits[s + 1]

        for i in samples_to_plot:

            mask = np.where( sample_array == i )[0]

            for j in mask:

                if file_data[j][0] == 3 or file_data[j][0] == 9:

                    hour_mask = np.where( (data[2] <= upper_bound) & (data[2] >= lower_bound) )[0]

                    plt.plot( data[2][hour_mask], data[1][j][[3, 6, 7, 8]][hour_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    shiny_de.append( data[2][hour_mask].tolist() )
                    shiny_de.append( data[1][j][[3, 6, 7, 8]][hour_mask].tolist() )

                else:

                    hour_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                    plt.plot( data[0][hour_mask], data[1][j][hour_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    shiny_de.append( data[0][hour_mask].tolist() )
                    shiny_de.append( data[1][j][hour_mask].tolist() )

        if ip.shiny:

            return shiny_de

        plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0, fontsize = 18 )
        # plt.legend( ncol = 2, fontsize = 18, bbox_to_anchor = ( 1.05, 1 ) )

        plt.xlabel( "Time [Hours]", fontsize = 18 )
        plt.ylabel( "Bottle Success Rate [%]", fontsize = 18 )
        plt.xticks( fontsize = 18 )
        plt.yticks( fontsize = 18 )

        plt.tight_layout()

        # For overall pipeline figure.

        # ax = plt.gca()
        # ax.get_legend().remove()
        # plt.xlabel( "" )
        # plt.ylabel( "" )
        # plt.tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        # plt.tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        if savefig:

            plt.savefig( ip.output_directory + "ESCR/Plots/Plot.pdf" )

        else:

            plt.show()

        plt.close()

        if data_extraction_bool:

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

    return shiny_de



# Imports.

from . import Preprocessing
from . import Utilities as util
from . import GCMS_plotting

# Main function definition.

def GCMS_Analysis_Main( ip ):
    '''Main function for GCMS data.'''

    file_data, data, first_derivative_data, second_derivative_data, third_derivative_data = [], [], [], [], []

    if ip.read_files:

        file_data, data = Preprocessing.read_files_and_preprocess( ip.directory, ip.data_directory, ip.merge_groups )

    if ip.write_csv:

        Preprocessing.write_csv( ip.output_directory, file_data, data )

    if ip.read_csv:

        file_data, data = Preprocessing.read_csv( ip.directory, ip.output_directory, ip.merge_groups )

    # Data of form [retention time, [Absolute intensity]]. Data of type lists of numpys.

    print( str( len( file_data ) ) + " files have been read." )

    if len( file_data ) != 12:

        print( "Warning: 12 files have not been read." )

    if ip.remove_files:

        Preprocessing.remove_files( file_data, data )

    if ip.compute_mean:

        Preprocessing.compute_mean( ip.output_directory, file_data, data )

    if ip.read_mean:

        Preprocessing.read_mean( ip.output_directory, data )

    # Data now of form [retention time, [Absolute intensity], [m]]. Data of type lists of numpys.

    if ip.derivative:

        first_derivative_data = util.compute_derivatives( data )
        second_derivative_data = util.compute_derivatives( first_derivative_data )
        third_derivative_data = util.compute_derivatives( second_derivative_data )

    if ip.extract_features:

        pass

        # util.extract_GCMS_features( output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data )

    if ip.read_and_analyse_features:

        pass

        # util.read_and_analyse_FTIR_features( directory, ip, file_data )

    if ip.plot_data:

        GCMS_plotting.plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data )

    if ip.sandbox:

        pass

        # util.sandbox( directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data )



# Imports.

import numpy as np
import pandas as pd
import math

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( directory, file_data, data ):

    df = pd.read_excel( directory + "GCMS/Raw_Data/Malodour Controls.xlsx", "TICs", header = None )

    file_data.append( [0, 0, "Virgin 4 Aug", ""] )
    file_data.append( [1, 0, "KW 4 Aug", ""] )
    file_data.append( [1, 1, "KW 22 Aug", ""] )
    file_data.append( [1, 2, "KW 4 Aug (Normalised)", ""] )

    data[0].append( df[0].to_numpy( dtype = np.float64 )[6:] )
    data[0].append( df[3].to_numpy( dtype = np.float64 )[6:] )
    data[0].append( df[6].to_numpy( dtype = np.float64 )[6:] )
    data[1].append( df[1].to_numpy( dtype = np.float64 )[6:] )
    data[1].append( df[4].to_numpy( dtype = np.float64 )[6:] )
    data[1].append( df[7].to_numpy( dtype = np.float64 )[6:] )
    data[2].append( df[2].to_numpy( dtype = np.float64 )[6:] )
    data[2].append( df[5].to_numpy( dtype = np.float64 )[6:] )
    data[2].append( df[8].to_numpy( dtype = np.float64 )[6:] )

def read_raw_data_file_2( data_directory, i, file_data, data ):

    filename = data_directory + str( i ) + ".csv"

    labels = ["HOC 8/3/24", "HOC 19/3/24", "HOC 10/4/24", "HOC 20/6/24", "HOC 2/5/24", "HOC 16/5/24"]

    with open( filename, 'r' ) as file:

        x, y, z = [], [], []

        lines = file.read().splitlines()

        for line in lines:

            a_list = line.split( "," )

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                x.append( list_of_floats[0] )
                y.append( list_of_floats[1] )
                z.append( list_of_floats[2] )

    data[0].append( np.array( x ) )
    data[1].append( np.array( y ) )

    file_data.append( [18, i - 1, labels[i - 1], ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    file_data, data = [], [[], []]

    # read_raw_data_file_1( directory, file_data, data )

    for i in range( 1, 7 ):

        read_raw_data_file_2( data_directory, i, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    data[0] = data[0][0]

    # Limonene normalisation

    maximum, minimum = [], []
    retention_time_shift = []

    rt_mask_limonene = np.where( (data[0] <= 4.3) & (data[0] >= 4) )[0]
    rt_mask_min = np.where( (data[0] <= 3.4) & (data[0] >= 2.4) )[0]

    for i in range( len( data[1] ) ):

        maximum.append( data[1][i][rt_mask_limonene].max() )
        minimum.append( data[1][i][rt_mask_min].min() )
        retention_time_shift.append( int( math.floor( (data[0][rt_mask_limonene][data[1][i][rt_mask_limonene].argmax()] - 4.18) * 1200 ) ) )

    maximum_array = np.array( maximum ).astype( np.float32 )
    minimum_array = np.array( minimum ).astype( np.float32 )

    normalised_headspaces = []

    for i in range( len( data[1] ) ):

        normalised_headspaces.append( 100 * (data[1][i] - minimum_array[i]) / (maximum_array[i] - minimum_array[i]) )

        if retention_time_shift[i] > 0:

            normalised_headspaces[i] = np.pad( normalised_headspaces[i][retention_time_shift[i]:], (0, retention_time_shift[i]) )

        if retention_time_shift[i] < 0:

            normalised_headspaces[i] = np.pad( normalised_headspaces[i][:retention_time_shift[i]], (-retention_time_shift[i], 0) )

    data[1].extend( normalised_headspaces )

    data[0] = data[0][120:-120]

    for i in range( len( data[1] ) ):

        data[1][i] = data[1][i][120:-120]

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    pass

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    labels = ["HOC 8/3/24", "HOC 19/3/24", "HOC 10/4/24", "HOC 20/6/24", "HOC 2/5/24", "HOC 16/5/24"]

    for i in range( 6 ):

        file_data.append( [18, i + 6, "Normalised " + labels[i], ""] )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    array = data[0][:, np.newaxis]

    for i in range( len( data[1] ) ):

        array = np.hstack( (array, data[1][i][:, np.newaxis]) )

    np.savetxt( output_directory + "GCMS/Condensed_Data/GCMS_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.6f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "GCMS/Condensed_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv file.'''

    file_data = []

    labels = ["HOC 8/3/24", "HOC 19/3/24", "HOC 10/4/24", "HOC 20/6/24", "HOC 2/5/24", "HOC 16/5/24"]
    labels.extend( ["Normalised " + labels[i] for i in range( 6 )] )

    df = pd.read_csv( output_directory + "GCMS/Condensed_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = df.iloc[i, 0]
        specimen = df.iloc[i, 1]

        file_data.append( [resin, specimen, labels[i], ""] )

    # file_data.append( [0, 0, "Virgin 4 Aug", ""] )
    # file_data.append( [1, 0, "KW 4 Aug", ""] )
    # file_data.append( [1, 1, "KW 22 Aug", ""] )
    # file_data.append( [1, 2, "KW 4 Aug (Normalised)", ""] )

    df = pd.read_csv( output_directory + "GCMS/Condensed_Data/GCMS_data" + name_appendage + ".csv", sep = ",", header = None )

    data = [df.iloc[:, 0].to_numpy( dtype = np.float64 ), []]

    for i in range( 1, len( df.columns ) ):

        data[1].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[1].pop( r )
        data[2].pop(r)

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    m = gu.sample_mean( file_data, data[1] )

    array = m[0][:, np.newaxis]

    for i in range( 1, len( m ) ):

        array = np.hstack( (array, m[i][:, np.newaxis]) )

    np.savetxt( output_directory + "GCMS/Condensed_Data/Means" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.6f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    m = []

    df = pd.read_csv( output_directory + "GCMS/Condensed_Data/Means" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.columns ) ):

        m.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    if len( data ) > 2: # A previous list of means is already appended.

            data[2] = m

    else:

        data.append( m )

    return data



# Imports.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, savefig = False ):

    resin_data = gu.get_list_of_resins_data( ip.directory )

    # For overall pipeline figure.

    # mpl.rcParams['lines.linewidth'] = 2

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    samples_to_plot = [16, 23]
    # samples_to_plot = samples_present

    specimens = True
    all_specimens = True
    specimen_mask_by_index = [0, 1, 3]
    specimen_mask = []

    mean = False

    deriv0 = True
    deriv1 = False
    deriv2 = False

    split = False
    num_splits = 28
    split_length = 20
    splits = [split_length * (i + 2) for i in range( num_splits )]
    # splits = [78, 82]

    step = 1

    if not split:

        # splits = [0, 600]
        splits = [0, 300]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        splits = ip.shiny_split
        all_specimens = False

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = False

    for s in range( len( splits ) - 1 ):

        data_extraction = []

        lower_bound, upper_bound = splits[s], splits[s + 1]

        for i in samples_to_plot:

            if specimens:

                mask = np.where( sample_array == i )[0]

                if not ip.shiny:

                    if all_specimens:

                        specimen_mask = [file_data[mask[j]][2] for j in range( len( mask ) )]

                    else:

                        specimen_mask = [file_data[mask[j]][2] for j in specimen_mask_by_index]

                for j in mask:

                     if file_data[j][2] in specimen_mask:

                        if deriv0:

                            strain_mask = np.where( (data[5][j] <= upper_bound) & (data[5][j] >= lower_bound) )[0]

                            # data_extraction.append( data[5][j][strain_mask] )
                            # data_extraction.append( data[4][j][strain_mask] )

                            plt.plot( data[5][j][strain_mask], data[4][j][strain_mask], label = file_data[j][2], color = colours[i] )

                            shiny_de.append( data[5][j][strain_mask].tolist() )
                            shiny_de.append( data[4][j][strain_mask].tolist() )

                        if deriv1:

                            strain_mask = np.where( (first_derivative_data[5][j] <= upper_bound) & (first_derivative_data[5][j] >= lower_bound) )[0]

                            plt.plot( first_derivative_data[5][j][strain_mask][::step], first_derivative_data[4][j][strain_mask][::step], label = file_data[j][2], color = colours[i] )

                        if deriv2:

                            strain_mask = np.where( (second_derivative_data[5][j] <= upper_bound) & (second_derivative_data[5][j] >= lower_bound) )[0]

                            plt.plot( second_derivative_data[5][j][strain_mask], second_derivative_data[4][j][strain_mask], label = file_data[j][2], color = colours[i] )

        if ip.shiny:

            return shiny_de

        # plt.legend( ncol = 2, borderaxespad = 0 )
        # plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )
        plt.legend( ncol = 2, loc = 'upper right', borderaxespad = 0, fontsize = 20 )
        # plt.legend()

        plt.xlabel( "Strain [%]", fontsize = 20 )
        plt.ylabel( "Stress [MPa]", fontsize = 20 )
        plt.xticks( fontsize = 20 )
        plt.yticks( fontsize = 20 )

        plt.tight_layout()

        # For overall pipeline figure.

        # ax = plt.gca()
        # ax.get_legend().remove()
        # plt.xlabel( "" )
        # plt.ylabel( "" )
        # plt.tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        # plt.tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        if savefig:

            plt.savefig( ip.output_directory + "TT/Plots/Plot.pdf" )

        else:

            plt.show()

        plt.close()

        if data_extraction_bool:

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

            # df = pd.DataFrame( data_extraction )
            #
            # df = df.transpose()
            #
            # df.to_csv( directory + "Output/Plot_Coords/Unnamed.csv" )

    return shiny_de



# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, resin_data, file_data, data ):

    with open( filename, 'r' ) as file:

        lines = file.readlines()

        linenumber = 0

        for line in lines:

            if linenumber < 1:

                resins = line.rstrip().split( "," )[1:]

                data[1] = [[] for i in resins]

                linenumber += 1
                continue

            if line.rstrip():

                a_list = line.rstrip().split( "," )

                data[0].append( float( a_list[0] ) )

                for ind, i in enumerate( a_list[1:] ):

                    data[1][ind].append( float( i ) )

            else:

                break

        for r in resins:

            file_data.append( [int( r ), 0, resin_data.loc[int( r )]["Label"] + ".{}".format( 0 ), ""] )

    data[0] = np.array( data[0] )

    for i in range( len( data[1] ) ):

        data[1][i] = np.array( data[1][i] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    file_data, data = [], [[], [], np.array( [24, 48, 72, 96] )]

    read_raw_data_file_1( data_directory + "ESCR.csv", resin_data, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    pass

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    pass

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    pass

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv file.'''

    return [], []

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[1].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    pass

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    pass



# Imports.

from . import Preprocessing
from . import Utilities as util
from . import FTIR_plotting
from .. import Global_Utilities as gu
import numpy as np
from scipy.signal import savgol_filter

# Main function definition.

def FTIR_Analysis_Main( ip ):
    '''Main function for FTIR data.'''

    file_data, data, m_peaks, y_peaks, mag_m_peaks, mag_y_peaks, m_peaks_array, y_peaks_array, first_derivative_data, second_derivative_data, third_derivative_data = [], [], [], [], [], [], [], [], [], [], []

    if ip.read_files:

        file_data, data = Preprocessing.read_files_and_preprocess( ip.directory, ip.data_directory, ip.merge_groups )

    if ip.write_csv:

        Preprocessing.write_csv( ip.output_directory, file_data, data )

    if ip.read_csv:

        file_data, data = Preprocessing.read_csv( ip.directory, ip.output_directory, ip.merge_groups )

    # Data of form [x, [y]]. data[0], data[1][i] all have length 1762. Data of type lists of numpys.

    gu.print_files_read( len( data[1] ), 572 )

    if ip.remove_files:

        Preprocessing.remove_files( file_data, data, "mnfrpaeubdtv" )

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        if s.find( "m" ) > -0.5:

            if file_data[i][0] == 40:

                array = np.array( data[1][i] )

                array = savgol_filter( array, 7, 3 )

                data[1][i] = array.tolist()

    if ip.compute_mean:

        Preprocessing.compute_mean( ip.output_directory, file_data, data )

    if ip.read_mean:

        Preprocessing.read_mean( ip.output_directory, data )

    # Data now of form [x, [y], [m]]. Data of type lists of numpys.

    if ip.derivative:

        first_derivative_data = util.compute_derivatives( data )
        second_derivative_data = util.compute_derivatives( first_derivative_data )
        third_derivative_data = util.compute_derivatives( second_derivative_data )

    if ip.extract_features:

        m_peaks, m_peaks_array, y_peaks, y_peaks_array = util.extract_FTIR_features( ip.output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, ip.peak_threshold_1, ip.peak_threshold_2, ip.peak_spacing, ip.peak_limit )

    if ip.read_and_analyse_features:

        util.read_and_analyse_FTIR_features( ip, file_data )

    if ip.plot_data:

        FTIR_plotting.plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, m_peaks, m_peaks_array, y_peaks, y_peaks_array )

    if ip.sandbox:

        util.sandbox( ip.directory, ip.output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data )



# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)" )

    resin = int( pattern.search( f ).groups()[0] )

    dict_specimen_removed = {4:[0], 10:[2], 18:[2], 2:[5, 6], 8:[5, 6], 13:[5, 6], 22:[5, 6], 23:[5, 6]} #Why: Strain all zeros. Or not present.

    sheets = ["Sheet1", "Sheet2", "Sheet3", "Sheet4", "Sheet5", "Sheet6", "Sheet7"]

    for specimen, s in enumerate( sheets ):

        if resin in dict_specimen_removed.keys():

            if specimen in dict_specimen_removed[resin]:

                continue

        df = pd.read_excel( filename, s )

        crop = 0

        strain = df.iloc[:, 5].tolist()

        for i, j in enumerate( strain ):

            if j > 0.5:

                crop = i
                break

        for i in range( len( df.columns ) ):

            data[i].append( df.iloc[:, i].to_numpy( dtype = np.float64 )[crop:] )

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], [], [], [], [], [], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_1 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    pass

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    pass

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    for i, f in enumerate( file_data ):

        array = np.array( data[0][i] )[:, np.newaxis]

        for j in range( 1, 7 ):

            array = np.hstack( (array, np.array( data[j][i] )[:, np.newaxis]) )

        np.savetxt( output_directory + "TT/Condensed_Data/Resin{}_{}_" + name_appendage + ".csv".format( f[0], f[1] ), array, delimiter = ",", fmt = "%.4f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "TT/File_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    file_data = []

    df = pd.read_csv( output_directory + "TT/File_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = df.iloc[i, 0]
        specimen = df.iloc[i, 1]

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    data = [[], [], [], [], [], [], []]

    filenames = sorted( [os.path.basename( path ) for path in glob.glob( output_directory + "TT/Condensed_Data/*" ) if ("_" + name_appendage + ".csv") in path], key = gu.sort_raw_files_3 )

    for f in filenames:

        df = pd.read_csv( output_directory + "TT/Condensed_Data/" + f, sep = ",", header = None )

        for i in range( len( df.columns ) ):

            data[i].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )

        for i in range( len( data ) ):

            data[i].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    pass

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    pass



def argument_of_maximum_closest_to_wavenumber( x, y, wavenumber ):
    '''Finds the index of the maximum that is closest to a wavenumber.'''

    first_deriv = gu.derivative( x, y )

    local_max = []

    for j in range( 1, len( first_deriv ) ):

        if first_deriv[j] >= 0 and first_deriv[j - 1] <= 0:

            local_max.append( j + 1 )

    closest_dist = 1000
    index = 0

    for j in range( len( local_max ) ):

        if abs( x[local_max[j]] - wavenumber ) < closest_dist:

            closest_dist = abs( x[local_max[j]] - wavenumber )
            index = local_max[j]

    return index

def file_names( directory ):
    '''Outputs "filenames" object.'''

    os.chdir( directory )

    filenames = [] # Each element will have form ["PCR X", filename, X, Repetition of sample]

    dir_1 = sorted( glob.glob( "*" ), key = sort_files )

    for i in range( len( dir_1 ) ):

        dir = dir_1[i] + "/Data point table"

        os.chdir( directory + dir )

        pcr = int( dir_1[i][4:] )

        dir_2 = sorted( glob.glob( "*" ) )

        for j in range( len( dir_2 ) ):

            if pcr == 5 and j == 8:

                    filenames.append( [dir_1[i], dir_2[j], pcr, j] )
                    filenames.append( [dir_1[i], dir_2[j], pcr, j + 1] )
                    filenames.append( [dir_1[i], dir_2[j], pcr, j + 2] )

            elif pcr == 20 and j == 8:

                    filenames.append( [dir_1[i], dir_2[j], pcr, j] )
                    filenames.append( [dir_1[i], dir_2[j], pcr, j + 1] )
                    filenames.append( [dir_1[i], dir_2[j], pcr, j + 2] )

            else:

                filenames.append( [dir_1[i], dir_2[j], pcr, j] )

    filenames.append( ["Additives", "PCR-20 repeat", 26, 0] )
    filenames.append( ["Additives", "PCR-20 repeat", 26, 1] )
    filenames.append( ["Additives", "PCR-20 repeat", 26, 2] )

    additives = ["No additive", "IrgaCycle ", "Accelothene G5", "TegoCycle 310", "Irganox 1010", "Irganox 1010 + Irgafos 168", "Vistamaxx 6102"]

    for ind, a in enumerate( additives ):

        filenames.append( ["Additives", a, 27 + 3 * ind, 0] )
        filenames.append( ["Additives", a, 27 + 3 * ind, 1] )
        filenames.append( ["Additives", a, 27 + 3 * ind, 2] )
        filenames.append( ["Additives", a, 28 + 3 * ind, 0] )
        filenames.append( ["Additives", a, 28 + 3 * ind, 1] )
        filenames.append( ["Additives", a, 28 + 3 * ind, 2] )
        filenames.append( ["Additives", a, 29 + 3 * ind, 0] )
        filenames.append( ["Additives", a, 29 + 3 * ind, 1] )
        filenames.append( ["Additives", a, 29 + 3 * ind, 2] )

    return filenames

def extract_file_data_from_filenames( filenames ):
    """Returns file_data object of form [PCR Sample, PCR Specimen, PCR Label, Description]."""

    file_data = []

    for f in filenames:

        if f[2] == 0:

            file_data.append( [f[2], f[3], "V{}.{}".format( 1, f[3] ), ""] )

        elif f[2] == 16:

            file_data.append( [f[2], f[3], "V{}.{}".format( 6, f[3] ), ""] )

        elif f[2] == 17:

            file_data.append( [f[2], f[3], "V{}.{}".format( 7, f[3] ), ""] )

        elif f[2] == 19:

            file_data.append( [f[2], f[3], "V{}.{}".format( 8, f[3] ), ""] )

        else:

            file_data.append( [f[2], f[3], "PCR{}.{}".format( f[2], f[3] ), ""] )

    return file_data

def read_file( filename, x, y ):
    '''Read file, extracting x, y values where x varies from 599 to 3997.'''

    with open( filename, 'r' ) as file:

        lines = file.readlines()

        for line in lines:

            a_list = line.split()

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                if list_of_floats[0] <= 3996.26214 and list_of_floats[0] >= 599.82506:

                    x.append( list_of_floats[0] )
                    y.append( list_of_floats[1] )

            else:

                break

def read_file_tf( filename, x1, y1, x2, y2, x3, y3 ):
    '''Read file, extracting x, y values where x varies from 599 to 3997.'''

    with open( filename, 'r' ) as file:

        lines = file.read().splitlines()

        counter = 0

        for line in lines:

            if counter < 2:

                counter += 1
                continue

            counter += 1

            a_list = line.split(",")

            a_list = [a_list[0], a_list[1], a_list[3], a_list[4], a_list[6], a_list[7]]

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                if list_of_floats[0] <= 3996.26214 and list_of_floats[0] >= 599.82506:

                    x1.append( list_of_floats[0] )
                    y1.append( list_of_floats[1] )
                    x2.append( list_of_floats[2] )
                    y2.append( list_of_floats[3] )
                    x3.append( list_of_floats[4] )
                    y3.append( list_of_floats[5] )

            else:

                break

def read_files( directory, filenames ):
    '''Reads the files. y is a list of arrays, an array for each file, x is just a single array.'''

    x, y = [], []

    for i in range( len( filenames ) ):

        if filenames[i][2] < 26:

            if not ((filenames[i][2] == 5 and filenames[i][3] > 7) or (filenames[i][2] == 20 and filenames[i][3] > 7)):

                x1, y1 = [], []

                filename = directory + filenames[i][0] + "/Data point table/" + filenames[i][1]

                read_file( filename, x1, y1 )

                x.append( np.array( x1 ) )
                y.append( np.array( y1 ) )

            else:

                if filenames[i][3] > 8:

                    continue

                else:

                    x1, y1, x2, y2, x3, y3 = [], [], [], [], [], []

                    filename = directory + filenames[i][0] + "/Data point table/" + filenames[i][1]

                    read_file_tf( filename, x1, y1, x2, y2, x3, y3 )

                    x.append( np.array( x1 ) )
                    x.append( np.array( x2 ) )
                    x.append( np.array( x3 ) )

                    y.append( np.array( y1 ) )
                    y.append( np.array( y2 ) )
                    y.append( np.array( y3 ) )

        else:

            if (filenames[i][2] == 26 or filenames[i][2] % 3 == 0) and (filenames[i][3] == 0):

                df = pd.read_excel( directory[:-14] + "TF035_FT-IR_raw data.xlsx", filenames[i][1] )

                if filenames[i][2] == 26:

                    df = df.drop( df.columns[[2, 5]], axis = 1 )
                    df = df.drop( [0], axis = 0 )

                    x.append( df[df.columns[0]].to_numpy() )
                    x.append( df[df.columns[2]].to_numpy() )
                    x.append( df[df.columns[4]].to_numpy() )

                    y.append( df[df.columns[1]].to_numpy() )
                    y.append( df[df.columns[3]].to_numpy() )
                    y.append( df[df.columns[5]].to_numpy() )

                elif filenames[i][2] == 45:

                    df = df.drop( df.columns[[2, 5, 8, 11, 14, 17, 20, 23]], axis = 1 )
                    df = df.drop( [0], axis = 0 )

                    array = df[df.columns[6]].to_list()

                    array_1 = []
                    array_2 = []

                    for i in array:

                        split = i.split( ',' )

                        array_1.append( float( split[0] ) )
                        array_2.append( float( split[1] ) )

                    df[df.columns[6]] = array_1
                    df[df.columns[7]] = array_2

                    x.append( df[df.columns[0]].to_numpy() )
                    x.append( df[df.columns[2]].to_numpy() )
                    x.append( df[df.columns[4]].to_numpy() )
                    x.append( df[df.columns[6]].to_numpy() )
                    x.append( df[df.columns[8]].to_numpy() )
                    x.append( df[df.columns[10]].to_numpy() )
                    x.append( df[df.columns[12]].to_numpy() )
                    x.append( df[df.columns[14]].to_numpy() )
                    x.append( df[df.columns[16]].to_numpy() )

                    y.append( df[df.columns[1]].to_numpy() )
                    y.append( df[df.columns[3]].to_numpy() )
                    y.append( df[df.columns[5]].to_numpy() )
                    y.append( df[df.columns[7]].to_numpy() )
                    y.append( df[df.columns[9]].to_numpy() )
                    y.append( df[df.columns[11]].to_numpy() )
                    y.append( df[df.columns[13]].to_numpy() )
                    y.append( df[df.columns[15]].to_numpy() )
                    y.append( df[df.columns[17]].to_numpy() )

                elif filenames[i][2] > 26:

                    df = df.drop( df.columns[[2, 5, 8, 11, 14, 17, 20, 23]], axis = 1 )
                    df = df.drop( [0], axis = 0 )

                    x.append( df[df.columns[0]].to_numpy() )
                    x.append( df[df.columns[2]].to_numpy() )
                    x.append( df[df.columns[4]].to_numpy() )
                    x.append( df[df.columns[6]].to_numpy() )
                    x.append( df[df.columns[8]].to_numpy() )
                    x.append( df[df.columns[10]].to_numpy() )
                    x.append( df[df.columns[12]].to_numpy() )
                    x.append( df[df.columns[14]].to_numpy() )
                    x.append( df[df.columns[16]].to_numpy() )

                    y.append( df[df.columns[1]].to_numpy() )
                    y.append( df[df.columns[3]].to_numpy() )
                    y.append( df[df.columns[5]].to_numpy() )
                    y.append( df[df.columns[7]].to_numpy() )
                    y.append( df[df.columns[9]].to_numpy() )
                    y.append( df[df.columns[11]].to_numpy() )
                    y.append( df[df.columns[13]].to_numpy() )
                    y.append( df[df.columns[15]].to_numpy() )
                    y.append( df[df.columns[17]].to_numpy() )

    return x, y

def extract_raw_data( directory ):

    filenames = file_names( directory + "FTIR/FTIR of rHDPE/" )

    file_data = extract_file_data_from_filenames( filenames )

    unstandard_x, y = read_files( directory + "FTIR/FTIR of rHDPE/", filenames )

    data = [unstandard_x, y]

    return file_data, data

def scale_and_shift( file_data, data ):
    '''Apply a scale and then a shift to each specimen dependent on how it relates to the mean for that sample.'''

    splits = [int( data[0][len( data[0] ) - 1] ), 800, 1400, 1500, 2000, 2700, 3000, int( data[0][0] )]

    m = gu.sample_mean( file_data, data[1] )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    for s in range( len( splits ) - 1 ):

        range_mask = np.where( (data[0] >= splits[s]) & (data[0] < splits[s + 1]) )

        for ind, i in enumerate( samples_present ):

            mask = np.where( sample_array == i )[0]

            for j in mask:

                scale_mean = 0

                for k in range( len( data[1][j][range_mask] ) ):

                    scale_mean += data[1][j][range_mask][k] / m[ind][range_mask][k]

                scale_mean /= len( data[1][j][range_mask] )

                data[1][j][range_mask] = data[1][j][range_mask] / scale_mean

    m = gu.sample_mean( file_data, data[1] )

    for s in range( len( splits ) - 1 ):

        range_mask = np.where( (data[0] >= splits[s]) & (data[0] < splits[s + 1]) )

        for ind, i in enumerate( samples_present ):

            mask = np.where( sample_array == i )[0]

            for j in mask:

                shift_mean = 0

                for k in range( len( data[1][j][range_mask] ) ):

                    shift_mean += m[ind][range_mask][k] - data[1][j][range_mask][k]

                shift_mean /= len( data[1][j][range_mask] )

                data[1][j][range_mask] = data[1][j][range_mask] + shift_mean

def find_index( frequency, x ):
    '''Finds the index in the list x that corresponds with the given wavenumber. (Was used for single frequency analysis.)'''

    for idx in range( len( x ) ):

        if x[idx] < frequency:

            if abs( x[idx] - frequency ) > abs( x[idx - 1] - frequency ):

                return idx - 1

            else:

                return idx

def difference_vectors( y ):
    '''For two arrays of the same length, returns an array where each entry is the difference between the two input arrays at that index.'''

    diffs = []

    for i in range( len( y ) ):

        diffs1 = []

        for j in range( len( y ) - 1 - i ):

            diffs1.append( np.array( y[i] ) - np.array( y[i + j + 1] ) )

        diffs.append( diffs1 )

    return diffs

def compute_vector_distances( dist_type, x, diffs, distance_matrix, arg_of_max ):
    '''Turn the difference vectors into simpler results.'''

    for i in range( len( diffs ) ):

        for j in range( len( diffs ) - 1 - i ):

            if dist_type == "max":

                distance_matrix[i][j + 1 + i] = distance_matrix[j + 1 + i][i] = np.absolute( diffs[i][j] ).max()

                for idx, k in enumerate( np.absolute( diffs[i][j] ) ):

                    if k == distance_matrix[i][j + 1 + i]:

                        arg_of_max[i][j + 1 + i] = arg_of_max[j + 1 + i][i] = x[idx]
                        break

            if dist_type == "sum":

                distance_matrix[i][j + 1 + i] = distance_matrix[j + 1 + i][i] = np.absolute( diffs[i][j] ).sum()

def compute_difference_matrix( output_directory, file_data, dist_type, data ):
    '''Compute difference matrices for sum and max.'''

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8]
    # sample_mask = samples_present

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    diffs = difference_vectors( data[1] )

    for dt in dist_type:

        max = False

        if dt == "max":

            max = True

        distance_matrix = np.zeros( (len( data[1] ), len( data[1] )) )
        arg_of_max = np.zeros( (len( data[1] ), len( data[1] )) )

        compute_vector_distances( dt, data[0], diffs, distance_matrix, arg_of_max )

        distance_matrix = distance_matrix[specimen_mask, :]
        distance_matrix = distance_matrix[:, specimen_mask]
        arg_of_max = arg_of_max[specimen_mask, :]
        arg_of_max = arg_of_max[:, specimen_mask]

        df = gu.array_with_column_titles_and_label_titles_to_df( distance_matrix, [f[2] for f in file_data_mask], [f[2] for f in file_data_mask] )

        df.to_csv( output_directory + "Matrices/" + dt + ".csv" )

        gu.plot_distance_matrix( output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "Matrices/", title = dt + ".pdf" )

        if max:

            df_arg = gu.array_with_column_titles_and_label_titles_to_df( arg_of_max, [f[2] for f in file_data_mask], [f[2] for f in file_data_mask] )

            df_arg.to_csv( output_directory + "Matrices/" + dt + "_arg.csv" )

            gu.plot_distance_matrix( output_directory, arg_of_max, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "Matrices/", title = dt + "_arg.pdf" )

def sigmoid_filter( magnitude, cutoff, threshold, shift, stretch ):
    '''Assigning a sigmoid filter to the magnitude of peaks.'''

    for i in range( len( magnitude ) ):

        for j in range( len( magnitude[i] ) ):

            if magnitude[i][j] < -cutoff:

                magnitude[i][j] = -1

            else:

                magnitude[i][j] = -1 / (1 + math.exp( -(stretch * ((2 * (-magnitude[i][j] - threshold) / (cutoff - threshold)) - 1) - shift) ))

# sigmoid_filter( mag_m_peaks, 0.00005, peak_threshold_1, 2, 5 )
# sigmoid_filter( mag_m_peaks, 0.00025, peak_threshold_1, 0, 3 )



def Plot_DSC_Curves( output_directory, Temperature, Heat_Flow, filenames ):
    '''Plot each DSC curve.'''

    os.chdir( output_directory + "Plots" )

    for i in range( len( filenames ) ):

        plt.plot( Temperature[2][i], Heat_Flow[2][i] )
        plt.plot( Temperature[4][i], Heat_Flow[4][i] )
        plt.title( filenames[i][:-4] )
        plt.savefig( filenames[i][:-4] + ".pdf" )
        plt.close()

# Code to compute whether two samples can be distinguished.

if distinguish_pairs and compute_max_var:

    distinguish_matrix = np.zeros( (num_samples, num_samples) )

    for s1 in range( len( samples_present ) ):

        for s2 in range( s1 + 1, len( samples_present ) ):

            mask_1 = np.where( sample == samples_present[s1] )[0]
            mask_2 = np.where( sample == samples_present[s2] )[0]

            if len( mask_1 ) < 3 or len( mask_2 ) < 3:

                continue

            distinguish = False
            i_val = 0

            for i in range( len( Heat_Flow[2][0] ) ):

                value_1 = np.zeros( len( mask_1 ) )
                value_2 = np.zeros( len( mask_2 ) )

                for j, k in enumerate( mask_1 ):

                    value_1[j] = Heat_Flow[2][k][i]

                for j, k in enumerate( mask_2 ):

                    value_2[j] = Heat_Flow[2][k][i]

                distinguish_at_point = True

                for j in value_1:

                    for k in value_2:

                        if abs( j - k ) < max_var_c[i]:

                            distinguish_at_point = False

                            break

                    if not distinguish_at_point:

                        break

                if distinguish_at_point:

                    distinguish = True
                    i_val = Temperature[2][0][i]
                    break

            if distinguish == True:

                print( "Samples " + str( samples_present[s1] ) + " and " + str( samples_present[s2] ) + " can be distinguished at temperature " + str( i_val ) )

                distinguish_matrix[s1][s2] = 1
                distinguish_matrix[s2][s1] = 1

            else:

                print( "Samples " + str( samples_present[s1] ) + " and " + str( samples_present[s2] ) + " cannot be distinguished." )

    os.chdir( output_directory + "Distinguish/" )

    distinguish_matrix = np.vstack( (samples_present_array[np.newaxis, :], distinguish_matrix) )

    nonzero_mask = np.nonzero( integral_c )[0]

    distinguish_matrix = distinguish_matrix[:, nonzero_mask]

    nonzero_mask = nonzero_mask + 1

    nonzero_mask = np.concatenate( [np.array( [0] ), nonzero_mask] )

    distinguish_matrix = distinguish_matrix[nonzero_mask, :]

    np.savetxt( "Distinguish.txt", distinguish_matrix )

def distinguish_distance( mask, list_1, list_2 ):

    match = 0

    for i in list_1:

        for j in list_2:

            if i == j:

                match = match + 1

                break

    for i in mask:

        if i not in list_1 and i not in list_2:

            match = match + 1

    return 1 - match / len( mask )

def read_and_analyse_distinguish_matrix( output_directory, filenames, plot_distinguish_dendrogram ):

    os.chdir( output_directory + "Distinguish/" )

    distinguish_matrix = np.loadtxt( "Distinguish.txt" )

    samples_present_array = distinguish_matrix[0, :]
    num_samples = len( samples_present_array )

    distinguish_matrix = distinguish_matrix[1:, :]

    differences_list = []

    for j, i in enumerate( distinguish_matrix ):

        a = (i != 0)
        b = list( samples_present_array[a] )

        differences_list.append( b )

    distance_matrix = np.zeros_like( distinguish_matrix )

    for i in range( len( differences_list ) - 1 ):

        for j in range( i + 1, len( differences_list ) ):

            distance_matrix[i][j] = distance_matrix[j][i] = distinguish_distance( samples_present_array, differences_list[i], differences_list[j] )

    if plot_distinguish_dendrogram:

        DSC_plotting.plot_distinguish_dendrogram( samples_present_array, distance_matrix )

def plot_distinguish_dendrogram( mask, distance_matrix ):

    condensed_distance_matrix = squareform( distance_matrix )
    linkage_matrix = linkage( condensed_distance_matrix, "single" )
    dendrogram( linkage_matrix, labels = mask )

    plt.ylim( [-0.02, 1] )
    plt.xlabel( "PCR Sample" )
    plt.title( "Differentiating samples given variation" )
    plt.savefig( "Distinguish.pdf" )

if read_distinguish_matrix:

    util.read_and_analyse_distinguish_matrix( output_directory, filenames, plot_distinguish_dendrogram )

bound_temperature = False
range_mask_c_u, range_mask_c_l = 155, 45 #155, 45. # Range of focus for crystallisation. For plot variance bar, used 130, 80. Else 195, 0.
range_mask_m_u, range_mask_m_l = 150, 100 # Range of focus for melt. For plot variance bar, used 180, 100. Else 205, 60.

if bound_temperature:

    Preprocessing.bound_temperature( Temperature, Heat_Flow, m, range_mask_c_u, range_mask_c_l, range_mask_m_u, range_mask_m_l )

def bound_range( temperature, upper_bound = 205, lower_bound = 0 ):
    '''Creates a mask with which to bound the data.'''

    mask = np.where( (temperature <= upper_bound) & (temperature >= lower_bound) )[0]

    return mask

def bound_temperature( Temperature, Heat_Flow, m, range_mask_c_u, range_mask_c_l, range_mask_m_u, range_mask_m_l ):
    '''Bound the data so that it lies within a specified temperature range.'''

    range_mask_c = bound_range( np.array( Temperature[2][0] ), range_mask_c_u, range_mask_c_l )
    range_mask_m = bound_range( np.array( Temperature[4][0] ), range_mask_m_u, range_mask_m_l )

    for i in range( len( Heat_Flow[2] ) ):

        Heat_Flow[2][i] = list( np.array( Heat_Flow[2][i] )[range_mask_c] )

    Temperature[2][0] = list( np.array( Temperature[2][0] )[range_mask_c] )

    for i in range( len( Heat_Flow[4] ) ):

        Heat_Flow[4][i] = list( np.array( Heat_Flow[4][i] )[range_mask_m] )

    Temperature[4][0] = list( np.array( Temperature[4][0] )[range_mask_m] )

    for i in range( len( m[0] ) ):

        m[0][i] = list( np.array( m[0][i] )[range_mask_c] )

    for i in range( len( m[1] ) ):

        m[1][i] = list( np.array( m[1][i] )[range_mask_m] )
